from dataclasses import dataclass

from yt_dlp import YoutubeDL

from spotidl.spotify import SpotifySong
from spotidl.utils import make_song_title


@dataclass
class YoutubeSong:
    id: str
    title: str
    video_url: str


def get_config(user_params: dict, song: SpotifySong) -> dict:
    """
    Prepares the parameters that need to be passed onto the YoutubeDL object.
    """

    downloader_params = {
        "postprocessors": [
            {
                "key": "FFmpegExtractAudio",
                "preferredcodec": user_params["codec"],
                "preferredquality": user_params["quality"],
            }
        ],
        "outtmpl": f"{make_song_title(song.artists, song.name, ', ')}.%(ext)s",
        # "outtmpl": "%(artist)s-%(title)s.ext",
        "quiet": user_params["quiet"],
        "format": "bestaudio/best",
        "dynamic_mpd": False,
    }

    return downloader_params


def get_downloader(params: dict):
    """
    Initiates the YoutubeDL class with the configured parameters.
    """

    return YoutubeDL(params=params)


def fetch_source(yt: YoutubeDL, song: SpotifySong) -> YoutubeSong:
    """
    Fetch apt source for the song from Youtube using the given details.
    """

    try:
        # using Song dataclasses' __repr__ function to construct the song name
        # adding "audio" to avoid 'official music videos' and similar types 😅
        song_title = make_song_title(song.artists, song.name, ", ") + " audio"
        search = yt.extract_info(f"ytsearch:{song_title}", download=False)
        yt_info = search["entries"][0]

    except Exception as e:
        print("Error when trying to get audio source from YT: ", e)
        return

    else:
        yt_song = YoutubeSong(
            id=yt_info["id"], title=yt_info["title"], video_url=yt_info["webpage_url"]
        )

        return yt_song


def download_song(yt: YoutubeDL, link: str):
    """
    Registers the provided parameters with the YoutubeDL object and
    downloads the song using the extracted information.
    """

    print("\nStarting song download...\n")

    try:
        # attempts to download the song using the best matched
        # youtube source link
        yt.download(link)

    except:
        print("\nDownload failed!")

    else:
        print(f"\nSuccessfully finished downloading!")


def controller(user_params: dict, song: SpotifySong):
    """
    Handles the flow of the download process for the given song.
    Initiates the configuration as per the user-defined parameters and chains
    the rest of functions together.
    """

    # user parameters are used in the downloader parameters dictionary
    # the downloader_params dict is then passed onto the YoutubeDL object
    # when generating its instance.
    downloader_params = get_config(user_params, song)
    yt = get_downloader(downloader_params)

    yt_song = fetch_source(yt, song)
    download_song(yt, yt_song.video_url)


if __name__ == "__main__":

    song = SpotifySong(
        name="He Don't Love Me",
        artists=["Winona Oak"],
        cover_url="",
        album_name="",
        disc_number=10,
        track_number=20,
    )

    user_params = {"codec": "mp3", "quality": "320", "quiet": True}

    controller(user_params, song)



from dotenv.main import load_dotenv
import spotipy
from spotipy.exceptions import SpotifyException
from spotipy.oauth2 import SpotifyOAuth

from dataclasses import dataclass

import spotidl.exceptions as e
import spotidl.utils as u

# loading .env vars
load_dotenv()

# check env vars
u.check_env_vars()

# initializing the spotify api connection
# the OAuth object automatically reads valid env. variables so we don't need to
# manually assign them using `os.environ.get(<env_var name>)`
try:
    sp = spotipy.Spotify(auth_manager=SpotifyOAuth())

except spotipy.oauth2.SpotifyOauthError as ex:
    # env variables aren't configured properly!
    raise e.EnvVariablesError("Environment variables arn't configured properly!")


# defining structure for the song data we are going to be parsing
@dataclass
class SpotifySong:
    name: str
    artists: dict
    album_name: str
    disc_number: int
    track_number: int
    cover_url: str

    def __str__(self):
        artists = []
        for artist in self.artists:
            artists.append(artist["name"])

        return u.make_song_title(artists, self.name, ", ")


def get_song_data(link: str) -> SpotifySong:
    """
    Gets relevant song details for the given Spotify song link.
    These are then passed onto other functions for further processing.
    """

    try:
        query = sp.track(link)

    except SpotifyException as ex:
        # wrapping the Spotify Exception
        # still unsure whether this is the correct approach
        raise e.NoDataReceivedError(ex)

    else:
        # there can be multiple results,
        # iterating over the list and extracting data
        artists = []
        for artist in query["artists"]:
            artists.append(artist["name"])

        song = new_song(query, type="song")

        return song


# this will serve as the common point for all entry types
# (individual song link, album link, etc.)
def new_song(query: dict, type: str, **album_details) -> SpotifySong:
    """
    Makes a new SpotifySong given a raw "track" type item received from Spotify API.
    The track queried using different methods returns slightly modified data.
    """

    artists = []
    for artist in query["artists"]:
        artists.append(artist["name"])

    name = u.correct_name(query["name"])

    if type == "song":
        # "song" refers to the case where the user enters a song link; we need
        # to fetch data for just a single song
        song = SpotifySong(
            name=name,
            artists=artists,
            album_name=query["album"]["name"],
            disc_number=query["disc_number"],
            track_number=query["track_number"],
            # the 1st link contains a link to the album art
            # of size 640 x 640 pixels
            cover_url=query["album"]["images"][0]["url"],
        )

    elif type == "album":
        song = SpotifySong(
            name=name,
            artists=artists,
            album_name=album_details["album_name"],
            disc_number=query["disc_number"],
            track_number=query["track_number"],
            cover_url=album_details["album_cover_url"],
        )

    return song


def get_album_data(link: str) -> tuple:
    """
    Gets album data for the given Spotify album link.
    It is then passed onto other functions for further processing.
    """

    # since our main data(all the relevant metadata, song URL) is associated
    # with the spotify song(referred to as "Track" in the spotify api), we
    # will need to make a list of all the songs in the album and then pass on
    # that list onto other functions.
    try:
        query = sp.album(link)

    except SpotifyException as ex:
        e.NoDataReceivedError(ex)

    album: list[SpotifySong] = []
    album_name = u.correct_name(query["name"])

    for track in query["tracks"]["items"]:
        # since this time we are directly receiving data that we otherwise
        # extracted from get_song_data using a link entered by the user, we
        # will need to generate a SpotifySong object another way

        song = new_song(
            track,
            type="album",
            album_name=album_name,
            album_cover_url=query["images"][0]["url"],
        )

        album.append(song)

    # returning album name since we'll need it when making the album folder to
    # use as the save directory
    return (album_name, album)


def get_playlist_data(link: str) -> tuple:
    """
    Gets playlist data for the given Spotify playlist link.
    It is then passed onto other functions for further processing.
    """

    # we need playlist id to fetch its information
    playlist_id = u.get_playlist_id(link)

    try:
        query = sp.playlist_tracks(playlist_id)
        playlist_name = sp.playlist(playlist_id)["name"]

    except SpotifyException as ex:
        raise e.NoDataReceivedError(ex)

    # can fetch a 100 tracks at a time
    # the "next" dictionary key gets the next batch, if the no. of results
    # exceeds 100
    tracks = query["items"]

    while query["next"]:
        query = sp.next(query)
        tracks.extend(query["items"])

    # now, to extract data from each entry in the list and get a SpotifySong object
    playlist_songs: list[SpotifySong] = []

    for track in tracks:
        song = new_song(track["track"], type="song")

        playlist_songs.append(song)

    return playlist_name, playlist_songs


if __name__ == "__main__":

    song = get_song_data(
        "https://open.spotify.com/track/0Ey8buiWgtBQjb7ypaACKN?si=22882b3a21ae4c71"
    )
    print(song)

    album = get_album_data(
        "https://open.spotify.com/album/0bWYlK9rRmIB68icHx9PNR?si=0PRvNMbBSticZvfdCeGdjw"
    )
    print(album[0])



import mutagen
import mutagen.easyid3
import mutagen.id3
import mutagen.flac
import mutagen.mp4
import mutagen.m4a

from spotidl.spotify import SpotifySong
from spotidl.utils import download_album_art


def add_metadata_mp3(file_name: str, song: SpotifySong, album_art_path: str):
    """
    Adds metadata to a MP3 file given it's path.
    File name must contain file extension as well.
    """

    # fix for invalid character in file name on windows
    if "?" in file_name:
        file_name = file_name.replace("?", "#")

    # writing textual metadata to the file
    try:
        meta = mutagen.easyid3.EasyID3(file_name)

    except mutagen.id3.ID3NoHeaderError:
        meta = mutagen.File(file_name, easy=True)
        meta.add_tags()

    meta["title"] = song.name
    #  ID3v2.3 spec: the delimiter is a "/" for multiple entries.
    meta["artist"] = "/".join(song.artists)
    meta["album"] = song.album_name
    meta["tracknumber"] = str(song.track_number)
    meta["discnumber"] = str(song.disc_number)

    meta.save(v2_version=3)

    # here we write the downloaded album art image to our downloaded song.
    meta = mutagen.id3.ID3(file_name)

    with open(album_art_path, "rb") as pic:
        meta["APIC"] = mutagen.id3.APIC(
            # type 3 is for the front cover of a song
            encoding=3,
            mime="image/jpeg",
            type=3,
            data=pic.read(),
        )

    # setting the v2 version to 3 is key- nothing will work without it
    meta.save(v2_version=3)


def add_metadata_flac(file_name: str, song: SpotifySong, album_art_path: str):
    """
    Adds metadata to a FLAC file given it's path.
    File name must contain file extension as well.
    """

    meta = mutagen.flac.FLAC(file_name)

    try:
        meta.add_tags()

    except mutagen.flac.FLACVorbisError:
        # if this error occurs, the vorbis comment block already exists
        meta.delete()

    meta["title"] = song.name
    meta["artist"] = "/".join(song.artists)
    meta["album"] = song.album_name
    meta["tracknumber"] = str(song.track_number)
    meta["discnumber"] = str(song.disc_number)

    # now, to write the album art to the file
    p = mutagen.flac.Picture()

    with open(album_art_path, "rb") as pic:
        p.data = pic.read()

    meta.add_picture(p)

    meta.save()


def add_metadata_m4a(file_name: str, song: SpotifySong, album_art_path: str):
    """
    Adds metadata to an M4A file given it's path.
    File name must contain file extension as well.
    """

    meta = mutagen.mp4.MP4(file_name)

    # refer https://stackoverflow.com/questions/56660170/\problem-using-mutagen-to-set-custom-tags-for-mp4-file

    meta["\xa9nam"] = song.name
    meta["\xa9ART"] = "/".join(song.artists)
    meta["\xa9alb"] = song.album_name

    # now, to write the album art
    with open(album_art_path, "rb") as pic:
        meta["covr"] = [
            mutagen.mp4.MP4Cover(
                pic.read(), imageformat=mutagen.mp4.MP4Cover.FORMAT_JPEG
            )
        ]

    meta.save(file_name)


def controller(file_name: str, song: SpotifySong, dir: str, codec: str):
    """
    Handles the metadata writing process flow.
    """

    album_art_path = download_album_art(
        path=dir, link=song.cover_url, title=song.album_name
    )

    print("\nWriting metadata...")

    if codec == "flac":
        add_metadata_flac(file_name, song, album_art_path)

    elif codec == "mp3":
        add_metadata_mp3(file_name, song, album_art_path)

    elif codec == "m4a":
        add_metadata_m4a(file_name, song, album_art_path)

    print("\nFinished writing metadata!")


# if __name__ == "__main__":
# add_cover(file_path="./yASD-dl/Rauw Alejandro, Tainy - ¿Cuándo Fue#")



# contains the custom exceptions to be used when attempting to enforce
# certain aspects in the application


class LinkError(Exception):
    """
    Raised when the provided link is incorrect.
    """

    # spits out whatever error message is passed to it
    pass


class NoDataReceivedError(Exception):
    """
    Raised when no data received from the external libraries.
    """

    pass


class EnvVariablesError(Exception):
    """
    Raised when environment variables aren't configured.
    """

    pass



import os
from re import search
from urllib.request import urlretrieve

import spotidl.config as c
import spotidl.exceptions as e


default_save_dir = os.getcwd() + "/spoti-dl"


def check_env_vars():
    """
    Run a barebones check to ensure that the three needed environment variables
    are not blank.
    """

    client_id, client_secret = os.getenv("SPOTIPY_CLIENT_ID"), os.getenv(
        "SPOTIPY_CLIENT_SECRET"
    )
    redirect_uri = os.getenv("SPOTIPY_REDIRECT_URI")

    if not client_id:
        raise e.EnvVariablesError("SPOTIPY_CLIENT_ID not configured!")

    elif not client_secret:
        raise e.EnvVariablesError("SPOTIPY_CLIENT_SECRET not configured!")

    elif not redirect_uri:
        raise e.EnvVariablesError("SPOTIPY_REDIRECT_URI not configured!")


# unused, will keep just in case *shrug*
def choice(msg: str) -> bool:
    """
    Offer the user a choice, infinite loop till they make a decision.
    """

    user_choice = ""
    choice_list = ("y", "n")
    while user_choice not in choice_list:
        user_choice = input(msg + ": ")

    return True if user_choice[0].lower() == "y" else False


def make_dir(path: str) -> bool:
    try:
        os.makedirs(path, exist_ok=True)

    except OSError as e:
        print("Error when attempting to make directory: ", e)

    else:
        return True


def check_dir(path: str) -> bool:
    return os.path.isdir(path)


def check_file(path: str) -> bool:
    return os.path.isfile(path)


def directory_maker(path: str):
    """
    Checks for the existence of a directory given the path
    and makes one if it doesn't exist.
    """

    if not check_dir(path):
        if make_dir(path):
            _, dir_name = os.path.split(path)
            print(f"Successfully created '{dir_name}' directory.")


if __name__ == "__main__":
    mk = directory_maker("./app/yASD-dl")
    print(mk)


def check_spotify_link(link: str, patterns_list: list) -> bool:
    """
    Handles all checks needed to vet user-entered Spotify links.
    """

    is_match = False

    if not link:
        return is_match

    # patterns_list contains a list of regex patterns for Spotify URLs
    for pattern in patterns_list:
        if search(pattern=pattern, string=link):
            is_match = True

    return is_match


def make_song_title(artists: list, name: str, delim: str):
    """
    Generates a song title by joining the song title and artist names.
    Artist names given in list format are split using the given delimiter.
    """

    return f"{delim.join(artists)} - {name}"


def download_album_art(
    path: str, link: str, title: str, extension: str = "jpeg"
) -> str:
    """
    Downloads album art- in a folder at the given path- to be embedded into songs.
    """

    # make a folder in the given path to store album art
    folder = "/album-art"
    full_path = path + folder
    directory_maker(full_path)

    download_path = full_path + f"/{title}.{extension}"

    if not check_file(download_path):
        urlretrieve(link, download_path)

    return download_path


def check_cli_args(codec: str, bitrate: str, link: str) -> bool:
    """
    Corrects audio-related arguments if they are incorrect and finally calls
    for Spotify link checks.
    """

    # adding checks to ensure argument validity
    if codec not in c.audio_formats:
        # raise argparse.ArgumentTypeError("Invalid codec entered!")
        print("Invalid codec entered! Using default value.")
        codec = "mp3"

    if bitrate not in c.audio_bitrates:
        # raise argparse.ArgumentTypeError("Invalid bitrate entered!")
        print("Invalid bitrate entered! Using default value.")
        bitrate = "320"

    # check whether the provided link is authentic
    is_match = check_spotify_link(link, c.spotify_link_patterns)

    return is_match


def get_link_type(link: str) -> str:
    """
    Analyzes the given link and returns the type of download type required.
    (track for individual song, album for an album URL and playlist if given
    a the Spotify link of a playlist).
    """

    # a spotify link is in the form of: open.spotify.com/<type>/<id>
    # we only need the type part
    type = link.split("/")

    return type[len(type) - 2]


def correct_name(query: str) -> str:
    """
    Checks for illegal characters in file or folder name and corrects them
    if needed.
    """

    for char in c.illegal_chars:
        if char in query:
            query = query.replace(char, "#")

    return query


# since we cannot fetch any playlist data without a playlist id, this will be
# of paramount importance
def get_playlist_id(link: str):
    """
    Returns a playlist ID given a Spotify playlist URL.
    """

    data = link.split("/")[-1]

    # we can safely return the last part if there's no question mark in it,
    # otherwise we need to split it again to remove the "si" code
    return data.split("?")[0] if "?" in data else data[-1]


if __name__ == "__main__":
    # testing

    # d = download_album_art(
    #     default_save_dir,
    #     "https://i.scdn.co/image/ab67616d0000b2730cf68eb8c1d1a406ba63162a",
    #     "Portrait For a Firewood",
    #     "jpeg",
    # )
    # print(d)

    type = get_link_type(
        "https://open.spotify.com/track/4Y9glyanYndWX8GNrRx1pI?si=947514736d6447d7"
    )
    print(type)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spotidl']

package_data = \
{'': ['*']}

install_requires = \
['python-dotenv>=0.19.2,<0.20.0',
 'spotipy>=2.19.0,<3.0.0',
 'yt-dlp>=2021.12.27,<2022.0.0']

entry_points = \
{'console_scripts': ['spotidl = spotidl.cli:controller']}

setup_kwargs = {
    'name': 'spoti-dl',
    'version': '0.9.6',
    'description': 'spotidl: download songs, albums and playlists using Spotify links',
    'long_description': '# Introduction\n\nspoti-dl(I had a better name but that was already taken on PyPi), is a song downloader app that accepts Spotify links, fetches individual song—and basic album—metadata from Spotify, downloads the song from Youtube. The metadata is then written onto the downloaded song file using the trusty Mutagen library, this includes the album/song cover art as well. \n\nThe app currently supports downloading songs, albums and playlists. \n\nI got the inspiration for the project from my friend Swapnil\'s [spotify-dl](https://github.com/SwapnilSoni1999/spotify-dl) app written in JavaScript. This seemed like the perfect pet project to make and consequently learn from :)\n\n\n# Setup\n\nRun ```pip install spoti-dl``` to install the app first and foremost.\n\nspoti-dl needs two things to work: [FFmpeg](https://ffmpeg.org/download.html) and a Spotify developer account.\n\nSteps to make a Spotify developer account:\n1. Go to [Spotify Dev Dashboard](https://developer.spotify.com/dashboard/applications)\n2. Login with your credentials and click on "create an app".\n3. Enter any name of choice, app description, tick the checkbox and proceed.\n4. Now you have access to your client ID. Click on "Show client secret" to get your client secret.\n5. From here, click on "edit settings" and in the "redirect URIs" section add any localhost URL. I personally use ```http://localhost:8080/callback```\n\nFinally, define these three environment variables: \n```\nSPOTIPY_CLIENT_ID\nSPOTIPY_CLIENT_SECRET\nSPOTIPY_REDIRECT_URI\n```\n\nAlso note that the first time you run the app you might get a popup window in your browser asking to integrate your account to the app you just created in the Spotify app dashboard. Accept and close the window.\n\n# Usage\n\nMP3 is the selected codec (audio file format; MP3 is the standard) and 320KBPS the selected bitrate (audio quality). So by default, running this downloads the song/album/playlist with these defaults-\n```\nspotidl <spotify link>\n``` \n\nas an example, running this would download Rick Astley\'s \'Never Gonna Give You Up\'- \n```\nspotidl https://open.spotify.com/track/4PTG3Z6ehGkBFwjybzWkR8?si=06f5d7ab5bd240e7\n```\n\nThe following audio formats are supported:\n- mp3 \n- flac\n- m4a (the m4a format only has partial textual-metadata support 😅))\n\nThe following bitrates are supported:\n- best \n- 320kbps\n- 256kbps \n- 192kbps (slightly better than Spotify\'s \'high\' audio setting, this is the bare minimum in my opinion to have a good listening experience)\n- 96kbps\n- 32kbps\n- worst\n\nAgain, the following link types are supported:\n- song links\n- album links\n- playlist links \n\nNote: File names (audio files or folder names (eg., playlist\'s directory name)) are changed to ensure compatibility with the operating systems since many characters like \'?\' or the \'/\' are illegal when making files/folders.',
    'author': 'Dhruv',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/good-times-ahead/spoti-dl/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# will contain all config options with multiple choices available;
# modifiable by command line arguments

# adding everything availble in the yt_dlp docs though i highly doubt
# anyone would want to download songs at anything below 128kbps
audio_bitrates = ["best", "320", "256", "192", "128", "96", "32", "worst"]

audio_formats = ["mp3", "flac", "m4a"]

spotify_link_types = ["track", "album", "playlist"]


# regex patterns for spotify URLs stored according to the url type
spotify_link_patterns = {
    "track": [
        r"^(spotify:|[a-z]+\.spotify\.com/track/)",
        r"^(spotify:|https://[a-z]+\.spotify\.com/track/)",
    ],
    "album": [
        r"^(spotify:|[a-z]+\.spotify\.com/album/)",
        r"^(spotify:|https://[a-z]+\.spotify\.com/album/)",
    ],
    # playlist will be implemented soon!
    "playlist": [
        r"^(spotify:|[a-z]+\.spotify\.com/playlist/)"
        r"^(spotify:|https://[a-z]+\.spotify\.com/playlist/)",
    ],
}

# list of illegal file name characters; replace them in case any of these
# is in the song name
illegal_chars = ["?", ">", "<", "|", ":", "/", "\\", "*"]


if __name__ == "__main__":
    from exceptions import *

    print("ok")



import argparse
import os
import pkg_resources

import spotidl.utils as u
import spotidl.spotify as s
import spotidl.youtube as y
import spotidl.config as c
import spotidl.exceptions as e
import spotidl.metadata as m


def cli_args():
    """
    Contains and parses all command line arguments for the application.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("link", help="Spotify song link to download")

    # important argument(s)
    parser.add_argument(
        "-d",
        "--dir",
        default=u.default_save_dir,
        help="Save directory(is created if doesn't exist)",
    )

    # audio-related arguments
    # quiet is a 'stored' argument, means we don't have to enter anything
    # after calling "-q/--quiet", it defaults to True if called else False
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Makes the downloader non-verbose/quiet",
    )

    parser.add_argument(
        "-c",
        "--codec",
        default="mp3",
        help=f"Audio format to download file as. List of available formats: {c.audio_formats}",
    )

    parser.add_argument(
        "-b",
        "--bitrate",
        default="320",
        help=f"Audio quality of the file. List of available qualities: {c.audio_bitrates}",
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=pkg_resources.get_distribution("spoti-dl").version,
        help="Displays the current app version",
    )

    # returns an argparse.Namespace object that stores our argument variables
    return parser.parse_args()


def controller():
    """
    Controls the flow of the program execution.
    """

    args = cli_args()

    #  perform necessary argument validity checks
    if not u.check_cli_args(args.codec, args.bitrate, args.link):
        raise e.LinkError("Invalid Spotify link entered!")

    # i believe getting the link type should be separated from just checking
    # the validity of the link and the audio-related args like codec and bitrate
    link_type = u.get_link_type(args.link)
    if not link_type in c.spotify_link_types:
        raise e.LinkError("Invalid Spotify link type entered!")

    # make the specified dir. if it doesn't exist and open it to store files
    u.directory_maker(args.dir)
    os.chdir(args.dir)

    # grouping all youtube-dl required arguments together before passing them
    # as the controller func parameters
    user_params = {
        "codec": args.codec,
        "quality": args.bitrate,
        "quiet": args.quiet,
        "dir": args.dir,
    }

    # replacing match-case with if else for comaptibility's sake
    if link_type == "track":
        song_download_controller(args.link, user_params)

    elif link_type == "album":
        # album controller requires the directory since it will
        # create a separate folder for the album
        album_download_controller(args.link, user_params)

    elif link_type == "playlist":
        playlist_download_controller(args.link, user_params)


def song_download_controller(link: str, user_params: dict):
    """
    Handles the control flow for the process to download an individual song.
    """

    # gets the SpotifySong dataclass object to be used for everything else in the func
    song = s.get_song_data(link)

    # use the youtube controller to scrape audio source and download the song
    y.controller(user_params, song)

    # write metadata to the downloaded file
    file_name = (
        f"{u.make_song_title(song.artists, song.name, ', ')}.{user_params['codec']}"
    )

    m.controller(file_name, song, user_params["dir"], user_params["codec"])


def album_download_controller(link: str, user_params: dict):
    """
    Handles the control flow for the process to download a complete album.
    """

    # get album information
    album_name, songs = s.get_album_data(link)
    save_dir = "./" + album_name

    # make a directory to store the album
    u.directory_maker(save_dir)
    os.chdir(save_dir)

    for song in songs:
        y.controller(user_params, song)

        # write metadata to the downloaded file
        file_name = (
            f"{u.make_song_title(song.artists, song.name, ', ')}.{user_params['codec']}"
        )

        # since we have already entered the <album_name> directory, we don't
        # have to pass in anything except the current directory indicator
        # we have to change directories since we are creating a dedicated folder
        # for the album unlike a song
        m.controller(file_name, song, ".", user_params["codec"])

    print(f"\nDownload for album '{album_name}' completed. Enjoy!")


def playlist_download_controller(link: str, user_params: dict):
    """
    Handles the control flow for the process to download a complete playlist.
    """

    # get playlist information
    playlist_name, songs = s.get_playlist_data(link)
    save_dir = "./" + playlist_name

    # make a directory to store the playlist
    u.directory_maker(save_dir)
    os.chdir(save_dir)

    for song in songs:
        y.controller(user_params, song)

        # write metadata to the downloaded file
        file_name = f"{u.make_song_title(song.artists, song.name, delim=', ')}.{user_params['codec']}"

        # passing "." aka curr dir indicator since we've already moved
        # into the <playlist_name> directory
        m.controller(file_name, song, dir=".", codec=user_params["codec"])

    print(f"\nDownload for playlist '{playlist_name}' completed. Enjoy!")
